#include <iostream>
#include <thread>
#include <unistd.h>
#include <signal.h>
#include <string>
#include <cstdlib>
#include <vector>
#include "Util.h"

using namespace std;

int n = 10;

void f(string val, int num) {
  cout << "[" << val << "] ----" << "\n";
  cout << "[" << val << "] num: " << num << "\n";
  cout << "[" << val << "] Thread id: " << this_thread::get_id() << "\n";
  if (num == 1) {
    std::thread t2(f, "t2", 0);
    std::thread t3(f, "t3", 0);
    t2.join();
    t3.join();
  }
  cout << "[" << val << "] " << n << "\n";
  n--;
  cout << "[" << val << "] ----" << "\n";
}

void SpawnProcesses(int n) {
  assert(n >= 0);
  vector<pid_t> pids;
  pids.reserve(n);
  Timer timer;

  // spawn the processes
  for (int i = 0; i < n; ++i) {
    timer.Tick();
    pid_t p = fork();
    if (p == 0) {
      // pause the child
      sleep(1000);
      exit(1);
    }
    timer.Tock();
    pids.push_back(p);

    if ((i + 1) % 10 == 0) {
      cout << i + 1 << "...\n";
    }
  }

  cout << "Creating " << n << " processes took " << timer.ElapsedTime() << " ms\n";
  cout << "Press return to kill the processes\n";
  cin.get();

  // kill the processes
  for (pid_t p : pids) {
    kill(p, SIGKILL);
  }
  cout << "Killed the processes\n";
}
void SpawnThreads(int n) {
  assert(n >= 0);
  vector<thread> threads;
  threads.reserve(n);
  Timer timer;

  // spawn the threads
  for (int i = 0; i < n; ++i) {
    timer.Tick();
    thread t([]() {
      std::this_thread::sleep_for(1000s);
    });
    timer.Tock();
    threads.push_back(move(t));

    if ((i + 1) % 10 == 0) {
      cout << i + 1 << "...\n";
    }
  }

  cout << "Creating " << n << " threads took " << timer.ElapsedTime() << " ms\n";
  cout << "Press return to finish the program\n";
  cin.get();

  // C++17 doesn't provide a way to stop threads, but there are options in
  // C++20. We will detach the threads and exit the program instead.
  for (thread& t : threads) {
    t.detach();
  }
}

int main(int argc, char * argv[]) {
  cout << "main thread id: " << this_thread::get_id() << "\n";
  std::thread t1(f, "one", 1); // add args to function here
  cout << "t1 id: " << t1.get_id() << "\n";
  cout << "Done." << "\n";

  t1.join();

  if (argc != 3) {
    cerr << "usage: " << argv[0] << " <#threads> <#processes>\n";
    return 1;
  }

  int n_threads = atoi(argv[1]);
  int n_processes = atoi(argv[2]);

  SpawnProcesses(n_processes);
  SpawnThreads(n_threads);
  
  return 0;
}
